/** A meaningfull atomic part of a table scheme definition */
class SchemeDefToken {
    constructor(
        content,        // Text content of token
        type,           // Type of token (see SchemeDefToken.TYPE_*)
        offset          // Position of token's start in source code
    ) {
        this.content = content;
        this.type = type;
        this.offset = offset;
    }

    /**
    * Create a human readable description of this token
    * @return String    Human readable description
    */
    toString() {
        switch (this.type) {
            case SchemeDefToken.TYPE_WHITESPACE:
                return 'Leerraum';
            case SchemeDefToken.TYPE_SEPARATOR:
                return 'Trennzeichen „' + this.content + '“';
            case SchemeDefToken.TYPE_UNDERSCORE:
                return 'Unterstreichungsumschalter';
            case SchemeDefToken.TYPE_IDENTIFIER:
                return 'Bezeichner „' + this.content + '“';
        }
        return '„' + this.content + '”';
    }
}

// Types of tokens
SchemeDefToken.TYPE_WHITESPACE = 0;
SchemeDefToken.TYPE_SEPARATOR = 1;
SchemeDefToken.TYPE_UNDERSCORE = 2;
SchemeDefToken.TYPE_IDENTIFIER = 3;

/**
* Convert string to an array of tokens
* @return Array     Array of SchemeDefToken instances
*/
SchemeDefToken.tokenize = function(
    strSchemeDef        // String describing table scheme
) {
    const tokens = [];
    const inputs = { s: strSchemeDef, pos: 0 };
    while (inputs.s.length) {
        const tokenData = [
            s => {
                const match = s.match(/^\s+/);
                return match ? { content: match[0], type: SchemeDefToken.TYPE_WHITESPACE } : null;
            },
            s => (s.charAt(0) == '_') ? { content: '_', type: SchemeDefToken.TYPE_UNDERSCORE } : null,
            s => ('.:,()[]'.indexOf(s.charAt(0)) != -1) ? { content: s.charAt(0), type: SchemeDefToken.TYPE_SEPARATOR } : null,
            s => {
                const match = s.match(/^[^\s:,\\(.\)\[\]]*[^\s:,\\(.\)\[\]_]/);
                return match ? { content: match[0], type: SchemeDefToken.TYPE_IDENTIFIER } : null;
            }
        ].reduce((prev, fkt) => (prev || fkt(inputs.s)), null);
        if (!tokenData)
            break;
        tokenData.length = tokenData.content.length;
        tokens.push(new SchemeDefToken(tokenData.content, tokenData.type, inputs.pos));
        inputs.pos += tokenData.length;
        inputs.s = inputs.s.substring(tokenData.length);
    }
    return tokens;
}
